"""
Track release upload history
"""
from cheddar.defaults import HISTORY_SIZE


class History(object):

    def __init__(self, app, size=HISTORY_SIZE):
        self.redis = app.redis
        self.size = size

    @property
    def key(self):
        return "cheddar.history"

    def add(self, name, version):
        """
        Add a new/version and truncate history.
        """
        self.redis.lpush(self.key, "{}/{}".format(name, version))
        self.redis.ltrim(self.key, 0, self.size - 1)

    def remove(self, name, version):
        """
        Remove a new/version from history.
        """
        self.redis.lrem(self.key, value="{}/{}".format(name, version))

    def all(self):
        """
        Return everything in the history.
        """
        return self.redis.lrange(self.key, 0, self.size - 1)

    def __len__(self):
        """
        Return size of history.
        """
        return self.redis.llen(self.key)



"""
Version and metadata utilities.
"""
from pkg_resources import parse_version

from pkginfo import SDist


def read_metadata(path):
    """
    Extract package metadata from source distribution.

    :param path: path to source distribution
    :returns: dict of meta data
    """
    distribution = SDist(path)
    return {key: getattr(distribution, key) for key in distribution.iterkeys()}


def sort_key(basename):
    """
    Define a sort key suitable for use in `sorted`
    that leverages the parsed version.
    """
    _, version = guess_name_and_version(basename)
    return parse_version(version)


def guess_name_and_version(basename):
    """
    Guess the distribution's name and version from its filename.
    """
    rest = basename

    for bad_extension in [".exe"]:
        if rest.endswith(bad_extension):
            # bad extension
            raise ValueError("Expected basename to have a valid package extension.")

    for extension in [".tar.gz", ".zip"]:
        if rest.endswith(extension):
            rest = rest[:- len(extension)]
            break

    name, version = rest.split("-", 1)
    if version[0].isdigit():
        # first split left
        return name, version
    else:
        # otherwise split right
        return rest.rsplit("-", 1)


def is_pre_release(basename):
    """
    Determine whether the version is a pre-release.
    """
    parsed_version = sort_key(basename)

    # check if the parsed version contains a non-numeric component
    #
    # 1.0.dev1 -> ('00000001', '00000000', '*@', '00000001', '*final')
    # 1.0c1    -> ('00000001', '00000000', '*c', '00000001', '*final')
    # 1.0      -> ('00000001', '00000000', '*final')
    # 1.0.1    -> ('00000001', '00000000', '00000001', '*final')
    try:
        [int(part) for part in parsed_version if part != "*final"]
        return False
    except ValueError:
        return True



"""
Implements distribution file storage.
"""
from os import makedirs, remove, walk
from os.path import basename, exists, isdir, join

from magic import from_buffer

from cheddar.model.versions import is_pre_release


class DistributionStorage(object):
    """
    File system storage with release/pre-release partitioning.
    """

    def __init__(self, base_dir, logger):
        """
        Initialize storage.

        :param base_dir: root directory for storage
        """
        self.logger = logger
        self.base_dir = base_dir
        self.release_dir = join(base_dir, "releases")
        self.pre_release_dir = join(base_dir, "pre-releases")
        self._make_base_dirs()

    def exists(self, name):
        return exists(self.compute_path(name))

    def read(self, name):
        """
        Read entry for name from storage.

        :returns: content data and content type, as a tuple
        """
        if not self.exists(name):
            self.logger.debug("No file exists for: {}".format(name))
            return None

        with open(self.compute_path(name)) as file_:
            content_data = file_.read()
            content_type = from_buffer(content_data, mime=True)
            self.logger.debug("Computed content type: {} for: {}".format(content_type, name))
            return content_data, content_type

    def write(self, name, data):
        """
        Write entry to storage.
        """
        self._make_base_dirs()
        path = self.compute_path(name)
        with open(path, "wb") as file_:
            file_.write(data)
        self.logger.debug("Wrote file for: {}".format(name))
        return path

    def remove(self, name):
        """
        Write entry to storage.
        """
        try:
            remove(self.compute_path(name))
            self.logger.debug("Removed file for: {}".format(name))
            return True
        except OSError:
            self.logger.debug("Unable to remove file for: {}".format(name))
            return False

    def compute_path(self, name):
        """
        Compute file system path.

        Path incorporates "pre-release" or "release" to easily
        differentiate released distributions for backup.
        """
        base_dir = self.pre_release_dir if is_pre_release(name) else self.release_dir
        path = join(base_dir, basename(name))
        self.logger.debug("Computed path: {} for: {}".format(path, name))
        return path

    def __iter__(self):
        for dirpath, _, filenames in walk(self.base_dir):
            for filename in filenames:
                yield join(dirpath, filename)

    def _make_base_dirs(self):
        """
        Ensure that base dirs exists.
        """
        for dir_ in [self.release_dir, self.pre_release_dir]:
            if not isdir(dir_):
                makedirs(dir_)



"""
URL mappings for package index functionality.
"""
from collections import OrderedDict
from functools import wraps
from urlparse import urljoin

from flask import abort, jsonify, make_response, render_template, request

from cheddar.auth import check_authentication
from cheddar.model.versions import sort_key


def create_routes(app):

    def authenticated(func):
        """
        Basic Auth decorator.
        """
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not check_authentication(app.redis):
                response = make_response("", 401)
                response.headers["WWW-Authenticate"] = 'Basic realm="cheddar"'
                return response
            return func(*args, **kwargs)
        return wrapper

    @app.route("/")
    def index():
        """
        Index page: show recent history.
        """
        app.logger.debug("Showing index page")
        return _render("index.html", history=app.history.all())

    @app.route("/simple/")
    @app.route("/simple")
    def list_project():
        """
        List all hosted projects.
        """
        app.logger.debug("Showing all projects")

        def name_sort_key(name):
            return name.lower()

        projects = sorted([project.name for project in app.index.get_projects()],
                          key=name_sort_key)

        return _render("simple.html", projects=projects)

    @app.route("/simple/<name>/")
    @app.route("/simple/<name>")
    def get_project(name):
        """
        List versions for a hosted project.
        """
        app.logger.info("Showing package index for: {}".format(name))

        versions = app.index.get_versions(name)

        if not versions:
            abort(404)

        sorted_versions = OrderedDict()
        for version in sorted(versions.keys(), key=sort_key, reverse=True):
            sorted_versions[version] = versions[version]

        return _render("project.html", project=name, versions=sorted_versions)

    @app.route("/simple/<name>/<version>/", methods=["GET", "DELETE"])
    @app.route("/simple/<name>/<version>", methods=["GET", "DELETE"])
    def handle_version(name, version):
        if request.method == "DELETE":
            return remove_version(name, version)
        else:
            return get_version(name, version)

    def get_version(name, version):
        """
        Get metadata for version.
        """
        app.logger.debug("Getting: {} {}".format(name, version))

        metadata = app.index.get_metadata(name, version)

        if metadata is None:
            abort(404)

        return _render("version.html", project=name, version=version, metadata=metadata)

    @authenticated
    def remove_version(name, version):
        """
        Delete all version data. Requires auth.
        """
        app.logger.debug("Removing: {} {}".format(name, version))
        app.index.remove_version(name, version)
        return ""

    @app.route("/local/<path:location>/")
    @app.route("/local/<path:location>")
    def get_local_distribution(location):
        """
        Local distribution download access.
        """
        app.logger.debug("Getting local distribution: {}".format(location))
        content_data, content_type = app.index.get_distribution(location, local=True)
        response = make_response(content_data)
        response.headers['Content-Type'] = content_type
        return response

    @app.route("/remote/<path:path>/")
    @app.route("/remote/<path:path>")
    def get_remote_distribution(path):
        """
        Remote distribution download access.

        Proxies and caches content.
        """
        # To account for redirects, we need to save the base url in the link URL along
        # with the path to the distribution. It's still a little awkward because the
        # urljoin logic happens here instead of within the remote index.
        location = urljoin(request.args["base"], path)

        app.logger.debug("Getting remote distribution: {}".format(location))
        content_data, content_type = app.index.get_distribution(location, local=False)
        response = make_response(content_data)
        response.headers['Content-Type'] = content_type
        return response

    @app.route("/pypi/", methods=["POST"])
    @app.route("/pypi", methods=["POST"])
    def pypi():
        """
        Register and upload endpoint for pip.
        """
        if request.files:
            return upload()
        else:
            return register()

    @authenticated
    def upload():
        """
        Upload a distribution. Requires auth.
        """
        app.logger.debug("Uploading distribution")

        _, upload_file = next(request.files.iterlists())
        app.index.upload_distribution(upload_file[0])

        return ""

    def register():
        """
        Register a distribution.

        The correct behavior for the `setuptools` register command is somewhat
        vague. The PEP 301 docs (for `distutils`) mostly cover user management,
        whereas `setuptools` CLI help says "register the distribution with the
        Python package index."

        Given this ambiguity and the fact that register does *NOT* send auth
        credentials, this operation simply validates the provided metadata.
        """
        app.logger.debug("Registering distribution")

        metadata = {key: values[0] for key, values in request.form.iterlists()}
        if not app.index.validate_metadata(metadata):
            abort(400)

        return ""

    def _render(template, **data):
        """
        Render response as either a template or just the raw JSON data.
        """
        if _wants_json():
            return jsonify(data)
        else:
            return render_template(template, **data)

    def _wants_json():
        """
        Should response use JSON?
        """
        for mime_type, _ in request.accept_mimetypes:
            # If we see a mime type that means HTML before JSON, return False
            if mime_type.startswith("text/html"):
                return False
            elif mime_type.startswith("application/xhtml"):
                return False
            # If we see JSON explicitly, return True
            elif mime_type.startswith("application/json"):
                return True
        # Default to False (HTML)
        return False



"""
Test version functions
"""
from os.path import dirname, join

from nose.tools import eq_

from cheddar.model.versions import (guess_name_and_version,
                                    is_pre_release,
                                    read_metadata,
                                    sort_key)


def test_parse_name_and_version():
    path = join(dirname(__file__), "../data/example-1.0.tar.gz")
    metadata = read_metadata(path)

    eq_(metadata,
        dict(author="Location Labs",
             author_email="info@locationlabs.com",
             description=None,
             home_page="http://www.locationlabs.com",
             keywords=None,
             license=None,
             metadata_version="1.0",
             name="example",
             platforms=["UNKNOWN"],
             summary="Example distribution",
             supported_platforms=(),
             version="1.0"))


def test_guess_name_and_version():

    def validate_guess(basename, expected_name, expected_version):
        name, version = guess_name_and_version(basename)
        eq_(name, expected_name)
        eq_(version, expected_version)

    cases = [("foo-1.0.tar.gz", "foo", "1.0"),
             ("foo-1.0c1.zip", "foo", "1.0c1"),
             ("foo-1.0.dev1", "foo", "1.0.dev1"),
             # hypenated names are a bad idea!
             ("foo-bar-1.0.tar.gz", "foo-bar", "1.0"),
             # weird versions extensions are also a bad idea (but may be required when forking)
             ("foo-1.0-bar.tar.gz", "foo", "1.0-bar")]
    for basename, expected_name, expected_version in cases:
        yield validate_guess, basename, expected_name, expected_version


def test_sort_key():
    versions = ["foo-1.1",
                "foo-1.0.1",
                "foo-1.0",
                "foo-1.0c1",
                "foo-1.0.dev10",
                "foo-1.0.dev9"]

    eq_(sorted(versions, key=sort_key),
        list(reversed(versions)))


def test_is_pre_release():

    def validate_is_pre_release(basename, expected):
        eq_(is_pre_release(basename), expected)

    cases = [("foo-1.1", False),
             ("foo-1.0.1", False),
             ("foo-1.0", False),
             ("foo-1.0c1", True),
             ("foo-1.0.dev10", True),
             ("foo-1.0.dev9", True)]
    for basename, expected in cases:
        yield validate_is_pre_release, basename, expected



"""
Factory entry point for the Flask application.

Allows development and deployment to configurate the Flask
instance differently.
"""
from flask import Flask

from cheddar.configure import configure_app


def create_app(debug=False, testing=False):
    """
    Create and configure the application.
    """
    app = Flask(__name__.split('.')[0])
    configure_app(app, debug=debug, testing=testing)
    return app



"""
Default configuration values.
"""
from getpass import getuser

FORCE_READ_REQUESTS = True

# Where do we get remote package data?
INDEX_URL = "http://pypi.python.org/simple"

# Where do we find Redis?
REDIS_HOSTNAME = 'localhost'

# How many seconds should we cache version content?
VERSIONS_TTL = 600

# How long should we wait for remote HTTP requests to complete?
# Note that "pip install" has a default timeout of 15 seconds...
GET_TIMEOUT = 20

# Where should we cache remote package data?
REMOTE_CACHE_DIR = "/var/tmp/cheddar-{}/remote".format(getuser())

# Where should we cache local package data?
LOCAL_CACHE_DIR = "/var/tmp/cheddar-{}/local".format(getuser())

# How much history to keep?
HISTORY_SIZE = 50

# Logging configuration
LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'debug': {
            'format': '%(asctime)s %(levelname)s: %(message)s [in %(pathname)s:%(lineno)d]',
            'datefmt': '%Y%m%d',
        },
        'default': {
            'format': '%(asctime)s - %(levelname)s - %(message)s',
            'datefmt': '%Y%m%d',
        },
    },

    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'debug',
            'stream': 'ext://sys.stdout',
        },
        'app': {
            'class': 'logging.handlers.TimedRotatingFileHandler',
            'level': 'DEBUG',
            'formatter': 'default',
            'filename': '/var/log/cheddar/cheddar.log',
        },
    },

    'loggers': {
        '': {
            'handlers': ['console', 'app'],
            'level': 'INFO',
            'propagate': False,
        },
    }

}



"""
Configure the Flask application.
"""
from logging.config import dictConfig
from time import ctime

from flask import request
from redis import Redis

from cheddar import defaults
from cheddar.controllers import create_routes
from cheddar.history import History
from cheddar.index.combined import CombinedIndex
from cheddar.index.storage import DistributionStorage
from cheddar.model.distribution import Projects


def configure_app(app, debug=False, testing=False):
    """
    Load configuration and initialize collaborators.
    """

    app.debug = debug
    app.testing = testing

    _configure_from_defaults(app)
    _configure_from_environment(app)
    _configure_logging(app)
    _configure_jinja(app)

    app.redis = Redis(app.config['REDIS_HOSTNAME'])
    app.projects = Projects(app.redis, app.logger)
    app.local_storage = DistributionStorage(app.config["LOCAL_CACHE_DIR"], app.logger)
    app.remote_storage = DistributionStorage(app.config["REMOTE_CACHE_DIR"], app.logger)
    app.history = History(app)
    app.index = CombinedIndex(app)

    if app.config.get('FORCE_READ_REQUESTS'):
        # read the request fully so that nginx and uwsgi play nice
        @app.after_request
        def read_request(response):
            request.stream.read()
            return response

    create_routes(app)


def _configure_from_defaults(app):
    """
    Load configuration defaults from defaults.py in this package.
    """
    app.config.from_object(defaults)


def _configure_from_environment(app):
    """
    Load configuration from a file specified as the value of
    the CHEDDAR_SETTINGS environment variable.

    Don't complain if the variable is unset.
    """
    app.config.from_envvar("CHEDDAR_SETTINGS", silent=True)


def _configure_logging(app):
    if app.debug or app.testing:
        app.config['LOGGING']['loggers']['']['handlers'] = ['console']
        if 'app' in app.config['LOGGING']['handlers']:
            del app.config['LOGGING']['handlers']['app']

    dictConfig(app.config['LOGGING'])


def _configure_jinja(app):
    def islist(obj):
        return isinstance(obj, list)

    def localtime(timestamp):
        return ctime(timestamp)

    app.jinja_env.filters.update({"islist": islist})
    app.jinja_env.filters.update({"localtime": localtime})



"""
Authentication.
"""
from flask import request


def check_authentication(redis):
    """
    Authenticate a request using a Redis lookup.
    """
    if request.authorization is None:
        return False

    key = "cheddar.user.{}".format(request.authorization.username)
    expected = redis.get(key)
    return expected is not None and expected == request.authorization.password



"""
Implements a combined local and remote package index.
"""
from cheddar.index.index import Index
from cheddar.index.local import LocalIndex
from cheddar.index.remote import CachedRemoteIndex


class CombinedIndex(Index):
    """
    Combined local and remote index.
    """

    def __init__(self, app):
        self.local = LocalIndex(app)
        self.remote = CachedRemoteIndex(app)
        self.logger = app.logger

    def get_projects(self):
        """
        Get projects from the local index.
        """
        return self.local.get_projects()

    def get_versions(self, name):
        """
        Get versions from both indexes, favoring the local index.
        """
        # A project hosted locally will mask anything hosted remotedly.
        # It *is* possible to query both indexes and merge the results, but
        # this comes as the cost of the added latency of the remote index
        # on every query with a cache miss.
        #
        # At the moment, this overhead doesn't seem worthwhile.
        local_versions = self.local.get_versions(name)
        if local_versions:
            self.logger.info("Obtained versions listing for: {} using local index".format(name))
            return local_versions

        remote_versions = self.remote.get_versions(name)
        self.logger.info("Obtained versions listing for: {} using remote index".format(name))
        return remote_versions

    def get_metadata(self, name, version):
        """
        Get metadata from local index.
        """
        return self.local.get_metadata(name, version)

    def get_distribution(self, path, **kwargs):
        """
        Get distribution using hint from controller.
        """
        local = kwargs.get("local", True)
        if local:
            return self.local.get_distribution(path, **kwargs)
        else:
            return self.remote.get_distribution(path, **kwargs)

    def remove_version(self, name, version):
        """
        Remove from local index.
        """
        self.local.remove_version(name, version)

    def validate_metadata(self, metadata):
        """
        Validate against the local index.
        """
        return self.local.validate_metadata(metadata)

    def upload_distribution(self, upload_file):
        """
        Upload to the local index.
        """
        return self.local.upload_distribution(upload_file)



"""
Implements a remote (proxy) package index.
"""
from json import dumps, loads
from os.path import abspath, join
from urllib import quote
from urlparse import urlsplit, urlunsplit

from BeautifulSoup import BeautifulSoup
from flask import abort
from requests import codes, get
from werkzeug.exceptions import HTTPException, NotFound

from cheddar.index.index import Index
from cheddar.model.versions import guess_name_and_version


class RemoteIndex(Index):
    """
    Access package data through a remote index server (e.g. pypi.python.org)
    """

    MAX_DEPTH = 2

    def __init__(self, app):
        self.index_url = app.config["INDEX_URL"]
        self.get_timeout = app.config["GET_TIMEOUT"]
        self.logger = app.logger

    def get_projects(self):
        """
        Unsupported.
        """
        raise NotImplementedError("get_projects")

    def get_versions(self, name):
        """
        Request version data from remote index and parse HTML.
        """
        url = "{}/{}".format(self.index_url, name)

        versions = {name: build_remote_path(href, location) for name, href, location in self._iter_listings(url, name)}

        self.logger.debug("Obtained remote version listing for: {}: {}".format(name, versions))
        return versions

    def get_metadata(self, name, version):
        """
        Unsupported.
        """
        pass

    def get_distribution(self, location, **kwargs):
        """
        Request distribution data for remote location.
        """
        self.logger.info("Getting remote distribution: {}".format(location))

        response = get(location, timeout=self.get_timeout)
        if response.status_code != codes.ok:
            self.logger.info("Distribution not found for: {}: {}".format(location, response.status_code))
            abort(codes.not_found)

        # don't log binary distribution content (.tar.gz, .zip, etc.), even at debug
        return response.content, response.headers["Content-Type"]

    def remove_version(self, name, version):
        """
        Unsupported.
        """
        raise NotImplementedError("remove_version")

    def validate_metadata(self, metadata):
        """
        Unsupported.
        """
        raise NotImplementedError("validate_metadata")

    def upload_distribution(self, file_):
        """
        Unsupported.
        """
        raise NotImplementedError("upload_distribution")

    def _iter_listings(self, url, name, depth=0):
        """
        Iterate through remote listings.

        Interpret version links and either yield (name, href, location) tuples
        or recursively spider to new links.
        """
        self.logger.info("Getting remote version listing for: {}".format(name))

        response = get(url, timeout=self.get_timeout)
        if response.status_code != codes.ok:
            self.logger.info("Remote version listing not found: {}".format(response.status_code))
            abort(codes.not_found)

        # Record the actual hostname used in case of redirection
        location = get_request_location(response, url)
        self.logger.debug("Index location was: {}".format(location))

        for link in iter_version_links(response.text, name):
            if isinstance(link, tuple):
                # Direct link
                name, href = link
                yield name, href, location
            else:
                # Recursive link
                if depth <= RemoteIndex.MAX_DEPTH:
                    try:
                        self.logger.info("Spidering to: {}".format(link))
                        for listing in self._iter_listings(link, name, depth + 1):
                            yield listing
                    except NotFound:
                        self.logger.debug("Unable to spider to: {}".format(link))
                else:
                    self.logger.info("Reached maximum depth; aborted spidering to: {}".format(link))


class CachedRemoteIndex(RemoteIndex):
    """
    Cache remote package data.

    - Since version data may change frequently, it is cached in Redis for simple expiration.
    - Distribution files are saved to the file system for easy inspection and backup.
    """
    def __init__(self, app):
        super(CachedRemoteIndex, self).__init__(app)
        self.redis = app.redis
        self.storage = app.remote_storage
        self.versions_ttl = app.config["VERSIONS_TTL"]
        self.logger = app.logger

    def get_versions(self, name):
        """
        Adds redis caching to versions listing.

        Currently, does not implement negative caching.
        """
        self.logger.info("Checking for cached versions listing for: {}".format(name))
        key = "cheddar.remote.{}".format(name)

        # Check cache
        cached_versions = self.redis.get(key)

        if cached_versions is not None:
            self.logger.debug("Found cached versions listing for: {}".format(name))
            return loads(cached_versions)

        try:
            computed_versions = super(CachedRemoteIndex, self).get_versions(name)
        except HTTPException as error:
            if error.code == codes.not_found:
                # Cache negative
                self.logger.debug("Caching negative versions listing for: {}".format(name))
                self.redis.setex(key, time=int(self.versions_ttl), value=dumps({}))
            else:
                self.logger.warn("Unexpected error querying remote versions listing", exc_info=True)
            raise
        else:
            self.logger.debug("Caching positive versions listing for: {}".format(name))
            self.redis.setex(key, time=int(self.versions_ttl), value=dumps(computed_versions))

        self.logger.debug(computed_versions)
        return computed_versions

    def get_distribution(self, location, **kwargs):
        """
        Cache distribution data.
        """
        cached = self.storage.read(location)
        if cached is not None:
            self.logger.debug("Found cached distribution for: {}".format(location))
            return cached

        content_data, content_type = super(CachedRemoteIndex, self).get_distribution(location, **kwargs)

        self.logger.debug("Caching distribution for: {}".format(location))
        self.storage.write(location, content_data)

        return content_data, content_type


def get_absolute_path(url, path):
    """
    Given a URL and a relative path, compute the URL's absolute path.
    """
    url_parts = list(urlsplit(url))
    return abspath(join(url_parts[2], path))


def get_base_url(url):
    """
    Given a URL, strip everything except the protocol and hostname.
    """
    url_parts = list(urlsplit(url))
    return urlunsplit(url_parts[:2] + [""] * 3)


def get_request_location(response, url):
    """
    Extract the request location from an HTTP response.
    """
    url_parts = urlsplit(url)
    # parse the request url
    scheme, netloc, path = url_parts.scheme, url_parts.netloc, url_parts.path

    # walk the list of redirects and update the url parts
    for redirect in response.history or [response]:
        if "location" not in redirect.headers:
            continue
        redirect_parts = urlsplit(redirect.headers["location"])
        if redirect_parts.scheme:
            scheme = redirect_parts.scheme
        if redirect_parts.netloc:
            netloc = redirect_parts.netloc
        if redirect_parts.path:
            path = redirect_parts.path

    # reconstruct the url (minus query string and fragments)
    return urlunsplit([scheme, netloc, path] + [""] * 2)


def build_remote_path(href, location):
    """
    Embed the index location into the link. If there was an #md5= fragment, it must
    come after the query string, so a little bit of gymnastics happens here.
    """
    if has_scheme(href):
        path = get_absolute_path(href, "")
        base_url = get_base_url(href)
    else:
        path = get_absolute_path(location, href)
        base_url = get_base_url(location)

    if "#" in path:
        base, fragment = path.split("#", 1)
        return "/remote{}?base={}#{}".format(base, quote(base_url, ""), fragment)
    else:
        return "/remote{}?base={}".format(path, quote(base_url, ""))


def has_scheme(url):
    """
    Does the input have a URL scheme?
    """
    try:
        url_parts = urlsplit(url)
        return bool(url_parts.scheme)
    except:
        return False


def iter_version_links(html, name):
    """
    Iterate through version links (in order) within HTML.

    Filtering out links that don't "look" like versions.

    Either yields hrefs to be recursively searches or tuples of (name, href)
    that match the given name.
    """
    soup = BeautifulSoup(html)
    for node in soup.findAll("a"):
        try:
            guessed_name, _ = guess_name_and_version(node.text)
        except ValueError:
            if node.get("rel") == "download":
                # Might be a recursive link.
                yield node["href"]
            # else couldn't parse name and version, probably the wrong kind of link
        else:
            if guessed_name != name:
                continue
            yield node.text, node["href"]



"""
Distribution model manipulation functions.
"""
from json import dumps, loads


class Projects(object):
    """
    Collection of hosted projects.
    """

    def __init__(self, redis, logger, prefix=""):
        self.redis = redis
        self.logger = logger
        self.prefix = prefix
        self.key = "{}cheddar.local".format(self.prefix)

    def list_projects(self):
        """
        Get all hosted projects.
        """
        return [Project(self, name) for name in self.redis.smembers(self.key)]

    def get_project(self, name):
        """
        Get a hosted project.
        """
        if self.redis.sismember(self.key, name):
            return Project(self, name)
        return None

    def add_project(self, name):
        """
        Add a hosted projects.
        """
        self.redis.sadd(self.key, name)
        return Project(self, name)

    def remove_project(self, name):
        """
        Remove a hosted projects.
        """
        self.redis.srem(self.key, name)

    def get_metadata(self, name, version):
        project = self.get_project(name)
        if project is None:
            return None

        project_version = project.get_version(version)
        if project_version is None:
            return None

        return project_version.get_metadata()

    def add_metadata(self, metadata):
        """
        Add a project, a version, and metadata.
        """
        name, version = metadata["name"], metadata["version"]

        self.logger.debug("Saving distribution: {} {}".format(name, version))

        project = self.add_project(name)
        project_version = project.add_version(version)
        project_version.set_metadata(metadata)

    def remove_metadata(self, name, version):
        """
        Remove metadata, version, and (maybe) project.
        """
        # Here be race conditions...
        project = Project(self, name)
        project_version = Version(project, version)

        project_version.remove_metadata()
        project.remove_version(version)

        if project.num_versions() == 0:
            project.remove()
            self.remove_project(name)


class Project(object):
    """
    A single hosted project.
    """
    def __init__(self, projects, name):
        self.redis = projects.redis
        self.logger = projects.logger
        self.prefix = projects.prefix
        self.name = name
        self.key = "{}cheddar.local.{}".format(self.prefix, self.name)

    def get_versions(self):
        """
        Get all versions for the project.
        """
        return [Version(self, version) for version in self.redis.smembers(self.key)]

    def num_versions(self):
        """
        Get the number of versions for a project.
        """
        return self.redis.scard(self.key)

    def get_version(self, version):
        """
        Add a version to a project.
        """
        if self.redis.sismember(self.key, version):
            return Version(self, version)
        return None

    def add_version(self, version):
        """
        Add a version to a project.
        """
        self.redis.sadd(self.key, version)
        return Version(self, version)

    def remove_version(self, version):
        """
        Remove a version from a project.
        """
        self.redis.srem(self.key, version)

    def remove(self):
        """
        Remove a project.
        """
        self.redis.delete(self.key)


class Version(object):
    """
    A single hosted project version.
    """

    FILENAME = "_filename"

    def __init__(self, project, version):
        self.redis = project.redis
        self.logger = project.logger
        self.prefix = project.prefix
        self.name = project.name
        self.version = version
        self.key = "{}cheddar.local.{}-{}".format(self.prefix, self.name, self.version)

    def get_metadata(self):
        """
        Get the version's metadata.
        """
        raw_metadata = self.redis.get(self.key)

        if raw_metadata is None:
            self.logger.debug("No metadata found for: {} {}".format(self.name, self.version))
            return None

        metadata = loads(raw_metadata)

        if Version.FILENAME not in metadata:
            self.logger.debug("Incomplete metadata for: {} {}".format(self.name, self.version))
            return None

        return metadata

    def set_metadata(self, metadata):
        """
        Set the version's metadata.
        """
        self.logger.debug("Saving metadata: {} for: {} {}".format(metadata, self.name, self.version))
        self.redis.set(self.key, dumps(metadata))

    def remove_metadata(self):
        """
        Remove the version's metadata.
        """
        self.redis.delete(self.key)



"""
Implements a local package index.
"""
from os.path import basename
from time import time

from flask import abort
from requests import codes
from werkzeug import secure_filename

from cheddar.index.index import Index
from cheddar.model.distribution import Version
from cheddar.model.versions import (guess_name_and_version,
                                    read_metadata)


class LocalIndex(Index):
    """
    Support register, upload, and management of locally hosted projects.
    """
    def __init__(self, app):
        self.redis = app.redis
        self.storage = app.local_storage
        self.logger = app.logger
        self.projects = app.projects
        self.history = app.history

    def get_projects(self):
        self.logger.info("Getting local projects")

        projects = self.projects.list_projects()

        self.logger.debug("Obtained local projects: {}".format(projects))
        return projects

    def get_versions(self, name):
        self.logger.info("Getting local versions listing for: {}".format(name))

        project = self.projects.get_project(name)
        if project is None:
            return None

        versions = {}
        for project_version in project.get_versions():
            metadata = project_version.get_metadata()
            if metadata is None:
                continue
            filename = metadata[Version.FILENAME]
            versions[filename] = "/local/{}".format(filename)

        self.logger.debug("Obtained local versions listing for: {}".format(name))
        return versions

    def get_metadata(self, name, version):
        self.logger.info("Getting local metatdata for: {} {}".format(name, version))

        metadata = self.projects.get_metadata(name, version)

        self.logger.debug("Obtained metadata: {} for: {}: {}".format(metadata, name, version))
        return metadata

    def get_distribution(self, location, **kwargs):
        self.logger.info("Getting local distribution: {}".format(location))

        result = self.storage.read(location)
        if result is None:
            self.logger.info("Distribution not found for: {}".format(location))
            abort(codes.not_found)

        # don't log binary version content (.tar.gz, .zip, etc.), even at debug
        return result

    def remove_version(self, name, version):
        """
        Remove redis and file data for project version.
        """
        self.logger.info("Removing version: {} {}".format(name, version))

        metadata = self.projects.get_metadata(name, version)
        if metadata is None:
            self.logger.info("Version not found: {} {}".format(name, version))
            abort(codes.not_found)

        self.storage.remove(metadata[Version.FILENAME])
        self.projects.remove_metadata(name, version)
        self.history.remove(name, version)

    def validate_metadata(self, metadata):
        """
        Validate that name and version are provided in the metadata.
        """
        self.logger.info("Validating metadata: {}".format(metadata))
        for required in ["name", "version"]:
            if required not in metadata:
                return False
        return True

    def upload_distribution(self, upload_file):
        """
        Upload distribution file and update redis data.
        """
        filename = secure_filename(upload_file.filename)
        self.logger.info("Uploading distribution: {}".format(filename))
        # don't log binary version content (.tar.gz, .zip, etc.), even at debug

        if self.storage.exists(filename):
            self.logger.warn("Aborting upload of: {}; already exists".format(filename))
            abort(codes.conflict)

        # write to storage first because read_metadata needs a file path
        path = self.storage.write(filename, upload_file.read())

        try:
            # extract metadata
            self.logger.debug("Parsing source distribution for metadata")
            metadata = self._get_metadata(path, filename)
        except:
            self.logger.debug("Removing uploaded file: {} on error".format(filename))
            self.storage.remove(filename)
            raise
        else:
            self.history.add(metadata["name"], metadata["version"])
            self.projects.add_metadata(metadata)

    def rebuild(self):
        """
        Rebuild redis index from file data.

        Rebuilding the index does not currently discard old index data,
        which could be accomplished by computing the set of all keys
        before and after the rebuild and removing the difference.
        """
        for path in self.storage:
            filename = basename(path)
            metadata = self._get_metadata(path, filename)
            self.projects.add_metadata(metadata)

    def _get_metadata(self, path, filename):
        metadata = read_metadata(path)

        # make sure metadata validates
        if not self.validate_metadata(metadata):
            abort(codes.bad_request)

        # make sure nothing fishy is going on
        if metadata.get(Version.FILENAME) not in [None, filename]:
            abort(codes.bad_request)

        # make sure metadata is consistent with filename
        expected_name, expected_version = guess_name_and_version(filename)
        if metadata["name"] != expected_name or metadata["version"] != expected_version:
            self.logger.warn("Aborting upload of: {}; conflicting filename and metadata".format(filename))
            abort(codes.bad_request)

        # include local path in metadata
        metadata[Version.FILENAME] = filename
        # add upload timestamp
        metadata["_uploaded_timestamp"] = time()

        return metadata



"""
Run a development server.
"""
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from cheddar.app import create_app


def main():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-a',
                        dest='all_interfaces',
                        action='store_true',
                        default=False,
                        help='Listen on all interfaces')
    parser.add_argument('-p',
                        dest='port',
                        type=int,
                        default=5000,
                        help='Listen port')
    args, extra = parser.parse_known_args()

    app = create_app(debug=True)

    if args.all_interfaces:
        app.run(host='0.0.0.0',
                port=args.port)
    else:
        app.run(port=args.port)



"""
Test remote index.
"""
from textwrap import dedent

from mock import MagicMock
from nose.tools import assert_raises, eq_

from cheddar.index.remote import (build_remote_path,
                                  get_absolute_path,
                                  get_base_url,
                                  get_request_location,
                                  iter_version_links)


def test_get_absolute_path():
    """
    Relative and absolute URL paths are converted to absolute paths.
    """
    eq_(get_absolute_path("http://foo.com/bar/baz", "../foo"), "/bar/foo")
    eq_(get_absolute_path("http://foo.com/bar/baz", "/foo"), "/foo")


def test_get_base_url():
    """
    Base URL preserves scheme, host, and port.
    """
    eq_(get_base_url("http://foo.com/bar/baz"), "http://foo.com")
    eq_(get_base_url("https://foo.com:443/foo/bar"), "https://foo.com:443")


def test_get_request_location_no_history_no_headers():
    """
    Request location defaults to the request url.
    """
    url = "http://pypi.python.org/simple/foo"
    response = MagicMock()
    response.history = []
    response.headers = {}
    eq_(get_request_location(response, url), url)


def test_get_request_location_no_history():
    """
    Request location follows response location header.
    """
    url = "http://pypi.python.org/simple/foo"
    response = MagicMock()
    response.history = []
    response.headers = {"location": "http://pypi.python.org/simple/foo/"}
    eq_(get_request_location(response, url), "http://pypi.python.org/simple/foo/")


def test_get_request_location_history():
    """
    Request location follows redirect history.
    """
    url = "http://pypi.python.org/simple/foo"

    redirect = MagicMock()
    redirect.headers = {"location": "https://pypi.python.org/simple/foo"}

    response = MagicMock()
    response.history = [redirect]
    response.headers = {}
    eq_(get_request_location(response, url), "https://pypi.python.org/simple/foo")


def test_get_request_location_multiple_history():
    """
    Request location follows redirect history across multiple redirects.
    """
    url = "http://pypi.python.org/simple/foo"

    redirect1 = MagicMock()
    redirect1.headers = {"location": "https://pypi.python.org/simple/foo"}

    redirect2 = MagicMock()
    redirect2.headers = {"location": "https://pypi.python.org/simple/foo/"}

    response = MagicMock()
    response.history = [redirect1, redirect2]
    response.headers = {}
    eq_(get_request_location(response, url), "https://pypi.python.org/simple/foo/")


def test_get_request_location_multiple_history_partial():
    """
    Request location follows redirect history across multiple redirects
    with partial paths.
    """
    url = "http://pypi.python.org/simple/foo-bar"

    redirect1 = MagicMock()
    redirect1.headers = {"location": "https://pypi.python.org/simple/foo-bar"}

    redirect2 = MagicMock()
    redirect2.headers = {"location": "/simple/foo-bar/"}

    redirect3 = MagicMock()
    redirect3.headers = {"location": "/simple/foo_bar"}

    redirect4 = MagicMock()
    redirect4.headers = {"location": "/simple/foo_bar/"}

    response = MagicMock()
    response.history = [redirect1, redirect2, redirect3, redirect4]
    response.headers = {}
    eq_(get_request_location(response, url), "https://pypi.python.org/simple/foo_bar/")


def test_build_remote_path():
    """
    Remote path incorporates path, base_url, and md5 fragmnt.
    """
    cases = [("../../packages/2.4/m/mock/mock-0.4.0-py2.4.egg#md5=e948e25b46f75d343a7fcdf24a36005c",
              "https://pypi.python.org/simple/mock/",
              "/remote/packages/2.4/m/mock/mock-0.4.0-py2.4.egg?base=https%3A%2F%2Fpypi.python.org#md5=e948e25b46f75d343a7fcdf24a36005c"),  # noqa
             ("http://effbot.org/media/downloads/PIL-1.1.7a2-py2.5-macosx10.5.mpkg.zip",
              "http://effbot.org/downloads/",
              "/remote/media/downloads/PIL-1.1.7a2-py2.5-macosx10.5.mpkg.zip?base=http%3A%2F%2Feffbot.org"),
             ("http://effbot.org/media/downloads/PIL-1.1.7.tar.gz",
              "http://effbot.org/downloads/",
              "/remote/media/downloads/PIL-1.1.7.tar.gz?base=http%3A%2F%2Feffbot.org")]

    def _validate(href, location, path):
        eq_(build_remote_path(href, location), path)

    for href, location, path in cases:
        yield _validate, href, location, path


def test_iter_version_links():
    """
    Versions links are correctly parsed.
    """
    HTML = dedent("""\
        <html>
          <body>
          <a href="../../packages/foo-1.0.tar.gz"/>foo-1.0.tar.gz</a>
          <a href="../../packages/bar-1.0.tar.gz"/>bar-1.0.tar.gz</a>
          <a href="http://foo.com/foo" rel="download"/>foo download link</a>
          </body>
        </html>""")

    iter_ = iter_version_links(HTML, "foo")
    eq_(next(iter_), ("foo-1.0.tar.gz", "../../packages/foo-1.0.tar.gz"))
    eq_(next(iter_), "http://foo.com/foo")

    with assert_raises(StopIteration):
        next(iter_)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

from cheddar import __version__

__build__ = ''

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(name='cheddar',
      version=__version__ + __build__,
      description='PyPI clone with Flask and Redis',
      long_description=readme + '\n\n' + history,
      author='Jesse Myers',
      author_email='jesse@locationlabs.com',
      url='https://github.com/jessemyers/cheddar',
      packages=find_packages(exclude=['*.tests']),
      setup_requires=[
          'nose>=1.3.0',
      ],
      install_requires=[
          'Flask>=0.10',
          'redis>=2.8.0',
          'requests>=2.0.0',
          'BeautifulSoup>=3.2.1',
          'python-magic>=0.4.6',
          'pkginfo>=1.1',
      ],
      tests_require=[
          'mock>=1.0.1',
          'mockredispy>=2.8.0.0',
      ],
      test_suite='cheddar.tests',
      entry_points={
          'console_scripts': [
              'development = cheddar.development:main',
          ]
      },
      include_package_data=True,
      )



"""
Abstraction around a package index.
"""
from abc import ABCMeta, abstractmethod


class Index(object):
    """
    Abstract interface for managing PyPI data.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_projects(self):
        """
        Get the list of hosted projects.

        :returns: an iterable of project names
        """
        pass

    @abstractmethod
    def get_versions(self, name):
        """
        Get the list of versions for a project.

        :param name: the project name
        :returns: a dictionary mapping versions to paths
        """
        pass

    @abstractmethod
    def get_metadata(self, name, version):
        """
        Get metadata for a project version.

        :param name: the project name
        :param version: the project version
        :returns: a dictionary of metadata
        """
        pass

    @abstractmethod
    def get_distribution(self, location, **kwargs):
        """
        Get distribution for a project version.

        :param location: location of the distribution content
        :returns: a pair of content data and content type for a distribution
        """
        pass

    @abstractmethod
    def remove_version(self, name, version):
        """
        Remove all data for a project version.

        :param name: the project name
        :param version: the project version
        """
        pass

    @abstractmethod
    def validate_metadata(metadata):
        """
        Validate a distribution's metadata.

        At a minimum, this function must verify that the project name and version
        are present in the dictionary. Other entries may be required as well.

        :param metadata: a dictionary of key-value metadata
        :returns: whether the metadata was valid
        """
        pass

    @abstractmethod
    def upload_distribution(self, upload_file):
        """
        Upload a distribution for a project version.

        The project name and version are determined from the metadata
        in the uploade_file contents.

        :param upload_file: an instance of `werkezeug.datastructures.FileStorage`
        """
        pass
